(self.webpackChunknglview_js_widgets=self.webpackChunknglview_js_widgets||[]).push([[776],{8516:(e,t,n)=>{(e.exports=n(9252)()).push([e.id,"",""])},9252:e=>{e.exports=function(){var e=[];return e.toString=function(){for(var e=[],t=0;t<this.length;t++){var n=this[t];n[2]?e.push("@media "+n[2]+"{"+n[1]+"}"):e.push(n[1])}return e.join("")},e.i=function(t,n){"string"==typeof t&&(t=[[null,t,""]]);for(var r={},o=0;o<this.length;o++){var i=this[o][0];"number"==typeof i&&(r[i]=!0)}for(o=0;o<t.length;o++){var s=t[o];"number"==typeof s[0]&&r[s[0]]||(n&&!s[2]?s[2]=n:n&&(s[2]="("+s[2]+") and ("+n+")"),e.push(s))}},e}},4246:e=>{var t={},n=function(e){var t;return function(){return void 0===t&&(t=e.apply(this,arguments)),t}},r=n((function(){return/msie [6-9]\b/.test(self.navigator.userAgent.toLowerCase())})),o=n((function(){return document.head||document.getElementsByTagName("head")[0]})),i=null,s=0,a=[];function f(e,n){for(var r=0;r<e.length;r++){var o=e[r],i=t[o.id];if(i){i.refs++;for(var s=0;s<i.parts.length;s++)i.parts[s](o.parts[s]);for(;s<o.parts.length;s++)i.parts.push(d(o.parts[s],n))}else{var a=[];for(s=0;s<o.parts.length;s++)a.push(d(o.parts[s],n));t[o.id]={id:o.id,refs:1,parts:a}}}}function u(e){for(var t=[],n={},r=0;r<e.length;r++){var o=e[r],i=o[0],s={css:o[1],media:o[2],sourceMap:o[3]};n[i]?n[i].parts.push(s):t.push(n[i]={id:i,parts:[s]})}return t}function l(e,t){var n=o(),r=a[a.length-1];if("top"===e.insertAt)r?r.nextSibling?n.insertBefore(t,r.nextSibling):n.appendChild(t):n.insertBefore(t,n.firstChild),a.push(t);else{if("bottom"!==e.insertAt)throw new Error("Invalid value for parameter 'insertAt'. Must be 'top' or 'bottom'.");n.appendChild(t)}}function c(e){e.parentNode.removeChild(e);var t=a.indexOf(e);t>=0&&a.splice(t,1)}function p(e){var t=document.createElement("style");return t.type="text/css",l(e,t),t}function d(e,t){var n,r,o;if(t.singleton){var a=s++;n=i||(i=p(t)),r=g.bind(null,n,a,!1),o=g.bind(null,n,a,!0)}else e.sourceMap&&"function"==typeof URL&&"function"==typeof URL.createObjectURL&&"function"==typeof URL.revokeObjectURL&&"function"==typeof Blob&&"function"==typeof btoa?(n=function(e){var t=document.createElement("link");return t.rel="stylesheet",l(e,t),t}(t),r=m.bind(null,n),o=function(){c(n),n.href&&URL.revokeObjectURL(n.href)}):(n=p(t),r=b.bind(null,n),o=function(){c(n)});return r(e),function(t){if(t){if(t.css===e.css&&t.media===e.media&&t.sourceMap===e.sourceMap)return;r(e=t)}else o()}}e.exports=function(e,n){if("undefined"!=typeof DEBUG&&DEBUG&&"object"!=typeof document)throw new Error("The style-loader cannot be used in a non-browser environment");void 0===(n=n||{}).singleton&&(n.singleton=r()),void 0===n.insertAt&&(n.insertAt="bottom");var o=u(e);return f(o,n),function(e){for(var r=[],i=0;i<o.length;i++){var s=o[i];(a=t[s.id]).refs--,r.push(a)}for(e&&f(u(e),n),i=0;i<r.length;i++){var a;if(0===(a=r[i]).refs){for(var l=0;l<a.parts.length;l++)a.parts[l]();delete t[a.id]}}}};var h,v=(h=[],function(e,t){return h[e]=t,h.filter(Boolean).join("\n")});function g(e,t,n,r){var o=n?"":r.css;if(e.styleSheet)e.styleSheet.cssText=v(t,o);else{var i=document.createTextNode(o),s=e.childNodes;s[t]&&e.removeChild(s[t]),s.length?e.insertBefore(i,s[t]):e.appendChild(i)}}function b(e,t){var n=t.css,r=t.media;if(r&&e.setAttribute("media",r),e.styleSheet)e.styleSheet.cssText=n;else{for(;e.firstChild;)e.removeChild(e.firstChild);e.appendChild(document.createTextNode(n))}}function m(e,t){var n=t.css,r=t.sourceMap;r&&(n+="\n/*# sourceMappingURL=data:application/json;base64,"+btoa(unescape(encodeURIComponent(JSON.stringify(r))))+" */");var o=new Blob([n],{type:"text/css"}),i=e.href;e.href=URL.createObjectURL(o),i&&URL.revokeObjectURL(i)}},309:(e,t,n)=>{var r=n(8516);"string"==typeof r&&(r=[[e.id,r,""]]),n(4246)(r,{}),r.locals&&(e.exports=r.locals)},3776:(e,t,n)=>{"use strict";n.r(t),n(309)}}]);