/* light theme */
/* text color */

* {
    color: #555555;
}

input,
textarea,
button,
select,
option {
    color: #111111;
}

input[type=range]:focus:after {
    color: #555555;
}

a{
    color: #1956d8;
}

input.Number, input.File {
    color: #79952e;
}


/* background color */

.OverlayPanel,
#menubar_ngl,
#menubar_ngl .menu .options,
#sidebar_ngl,
#toolbar_ngl
{
    background-color: #e1e4eb;
}


/* border color */

.OverlayPanel,
#menubar_ngl .menu .options hr,
#sidebar_ngl > .Panel,
#sidebar_ngl > .Content > .Panel {
    border-color: #cccccc;
}


/* special */

#menubar_ngl .menu .options .option:hover,
.option:hover > .Icon,
.option:hover > .Text {
    color: #D8D8D8;
    background-color: #345a69;
}

.highlight,
.highlight > .Icon,
.highlight > .Text {
    color: #D8D8D8;
    background-color: #345a69;
}

.EllipsisMultilineText:after {
    background: linear-gradient(to right, rgba(255, 255, 255, 0), #e1e4eb 50%, #e1e4eb);
}
